"""Creation and fetching of feed data."""
import threading
from typing import Any, Dict, Callable, Optional

import requests
from dataclasses import InitVar, dataclass

from routemaster.utils import get_path, template_url


def feeds_for_state_machine(state_machine) -> Dict[str, 'Feed']:
    """Get a mapping of feed prefixes to unfetched feeds."""
    return {
        x.name: Feed(x.url, state_machine.name)  # type: ignore
        for x in state_machine.feeds
    }


class FeedNotFetched(Exception):
    """Raised if we try to access the data in a feed we haven't yot fetched."""
    pass


_feed_sessions = threading.local()


def _get_feed_session():
    # We cache sessions per thread so that we can use `requests.Session`'s
    # underlying `urllib3` connection pooling.
    if not hasattr(_feed_sessions, 'session'):
        _feed_sessions.session = requests.Session()
    return _feed_sessions.session


@dataclass
class Feed:
    """A feed fetcher, able to retreive a feed and read keys out of it."""
    url: str
    state_machine: str
    data: InitVar[Optional[Dict[str, Any]]] = None

    def prefetch(
        self,
        label: str,
        log_response: Callable[[requests.Response], None] = lambda x: None,
    ) -> None:
        """Trigger the fetching of a feed's data."""
        if self.data is not None:
            return

        url = template_url(self.url, self.state_machine, label)

        session = _get_feed_session()
        response = session.get(url)
        log_response(response)
        response.raise_for_status()
        self.data = response.json()

    def lookup(self, path):
        """Lookup data from a feed's contents."""
        if self.data is None:
            raise FeedNotFetched(self.url)
        return get_path(path, self.data)



"""WSGI middlewares used in routemaster."""

from typing import Any, Dict, List, Callable, Iterable, Optional

from routemaster.app import App
from routemaster.utils import WSGICallable, StartResponse, WSGIEnvironment

WSGIMiddleware = Callable[[App, WSGICallable], WSGICallable]

ACTIVE_MIDDLEWARES: List[WSGIMiddleware]
ACTIVE_MIDDLEWARES = []


def middleware(fn: WSGIMiddleware):
    """Decorator: add `fn` to ACTIVE_MIDDLEWARES."""
    ACTIVE_MIDDLEWARES.append(fn)
    return fn


def wrap_application(app: App, wsgi: WSGICallable) -> WSGICallable:
    """Wrap a given WSGI callable in all active middleware."""
    for middleware_instance in reversed(ACTIVE_MIDDLEWARES):
        wsgi = middleware_instance(app, wsgi)
    return wsgi


# Definitions start here


@middleware
def session_middleware(app: App, wsgi: WSGICallable) -> WSGICallable:
    """Manage an ORM session around each request."""
    def inner(
        environ: WSGIEnvironment,
        start_response: StartResponse,
    ) -> Iterable[bytes]:
        def wrapped_start_response(
            status: str,
            headers: Dict[str, str],
            exc_info: Optional[Any] = None,
        ) -> None:
            start_response(status, headers, exc_info)
            if exc_info is not None:  # pragma: no branch
                app.set_rollback()  # pragma: no cover

        with app.new_session():
            yield from wsgi(environ, wrapped_start_response)
    return inner


@middleware
def logging_middleware(app: App, wsgi: WSGICallable) -> WSGICallable:
    """Log requests as they come in."""
    def inner(
        environ: WSGIEnvironment,
        start_response: StartResponse,
    ) -> Iterable[bytes]:

        kwargs: Dict[str, Any] = {}

        def wrapped_start_response(
            status: str,
            headers: Dict[str, str],
            exc_info: Optional[Any] = None,
        ) -> None:
            kwargs['status'] = status.split()[0]
            kwargs['headers'] = headers
            kwargs['exc_info'] = exc_info
            start_response(status, headers, exc_info)

        app.logger.process_request_started(environ)
        yield from wsgi(environ, wrapped_start_response)
        app.logger.process_request_finished(environ, **kwargs)

    return inner



"""Routemaster logging interface for Python's logging library."""

import time
import logging
import contextlib

from routemaster.logging.base import BaseLogger


class PythonLogger(BaseLogger):
    """Routemaster logging interface for Python's logging library."""

    def __init__(self, *args, log_level: str) -> None:
        super().__init__(*args)

        logging.basicConfig(
            format=(
                "[%(asctime)s] [%(process)d] [%(levelname)s] "
                "[%(name)s] %(message)s"
            ),
            datefmt="%Y-%m-%d %H:%M:%S %z",
            level=getattr(logging, log_level),
        )
        self.logger = logging.getLogger('routemaster')
        self.logger.info(f"Started logger with level {log_level}")

    @contextlib.contextmanager
    def process_cron(self, state_machine, state, fn_name):
        """Process a cron job, logging information to the Python logger."""
        self.logger.info(
            f"Started cron {fn_name} for state {state.name} in "
            f"{state_machine.name}",
        )
        try:
            time_start = time.time()
            yield
            duration = time.time() - time_start
        except Exception:
            self.logger.exception(f"Error while processing cron {fn_name}")
            raise

        self.logger.info(
            f"Completed cron {fn_name} for state {state.name} "
            f"in {state_machine.name} in {duration:.2f} seconds",
        )

    def process_request_finished(
        self,
        environ,
        *,
        status,
        headers,
        exc_info,
    ):
        """Process a web request and log some basic info about it."""
        self.info("{method} {path} {status}".format(
            method=environ.get('REQUEST_METHOD'),
            path=environ.get('PATH_INFO'),
            status=status,
        ))

    def __getattr__(self, name):
        """Fall back to the logger API."""
        return getattr(self.logger, name)



"""Public API for state machines."""

from routemaster.state_machine.api import (
    LabelRef,
    LabelProvider,
    LabelStateProcessor,
    list_labels,
    create_label,
    delete_label,
    process_cron,
    get_label_state,
    get_label_metadata,
    update_metadata_for_label,
)
from routemaster.state_machine.gates import process_gate
from routemaster.state_machine.utils import (
    labels_in_state,
    labels_in_state_with_metadata,
    labels_needing_metadata_update_retry_in_gate,
)
from routemaster.state_machine.actions import process_action
from routemaster.state_machine.exceptions import (
    DeletedLabel,
    UnknownLabel,
    LabelAlreadyExists,
    UnknownStateMachine,
)

__all__ = (
    'LabelRef',
    'list_labels',
    'create_label',
    'delete_label',
    'process_cron',
    'process_gate',
    'DeletedLabel',
    'UnknownLabel',
    'LabelProvider',
    'process_action',
    'get_label_state',
    'labels_in_state',
    'get_label_metadata',
    'LabelAlreadyExists',
    'LabelStateProcessor',
    'UnknownStateMachine',
    'update_metadata_for_label',
    'labels_in_state_with_metadata',
    'labels_needing_metadata_update_retry_in_gate',
)



"""Shared text utilities."""
from typing import Iterable


def join_comma_or(items: Iterable[str]) -> str:
    """Join strings with commas and 'or'."""
    if not items:
        raise ValueError("No items to join")

    *rest, last = items

    if not rest:
        return last

    return f"{', '.join(rest)} or {last}"



"""Action (webhook invocation) evaluator."""

import json
import hashlib
import functools

from routemaster.db import History
from routemaster.app import App
from routemaster.utils import template_url
from routemaster.config import State, Action, StateMachine
from routemaster.webhooks import WebhookResult
from routemaster.state_machine.types import LabelRef
from routemaster.state_machine.utils import (
    choose_next_state,
    context_for_label,
    get_label_metadata,
    get_current_history,
)
from routemaster.state_machine.exceptions import DeletedLabel


def process_action(
    *,
    app: App,
    state: State,
    state_machine: StateMachine,
    label: LabelRef,
) -> bool:
    """
    Process an action for a label.

    Assumes that `action` is the current state of the label, and that the label
    has been locked.

    Returns whether the label progressed in the state machine, for which `True`
    implies further progression should be attempted.
    """
    if not isinstance(state, Action):  # pragma: no branch
        raise ValueError(  # pragma: no cover
            f"process_action called with {state.name} which is not an Action",
        )

    action = state

    metadata, deleted = get_label_metadata(app, label, state_machine)
    if deleted:
        raise DeletedLabel(label)

    latest_history = get_current_history(app, label)

    webhook_data = json.dumps({
        'metadata': metadata,
        'label': label.name,
    }, sort_keys=True).encode('utf-8')

    run_webhook = app.get_webhook_runner(state_machine)

    idempotency_token = _calculate_idempotency_token(label, latest_history)

    webhook_logger = functools.partial(
        app.logger.webhook_response,
        state_machine,
        state,
    )

    with app.logger.process_webhook(state_machine, state):
        result = run_webhook(
            template_url(action.webhook, state_machine.name, label.name),
            'application/json',
            webhook_data,
            idempotency_token,
            webhook_logger,
        )

    if result != WebhookResult.SUCCESS:
        return False

    context = context_for_label(
        label,
        metadata,
        state_machine,
        action,
        latest_history,
        app.logger,
    )
    next_state = choose_next_state(state_machine, action, context)

    app.session.add(History(
        label_state_machine=state_machine.name,
        label_name=label.name,
        old_state=action.name,
        new_state=next_state.name,
    ))

    return True


def _calculate_idempotency_token(label: LabelRef, latest_history) -> str:
    """
    We want to make sure that an action is only performed once.

    While we attempt to only deliver an action webhook once, we cannot
    guarantee this in all cases, so we call webhooks with an
    _idempotency token_. This token allows the receiver to record that it has
    performed the appropriate action, and should not perform it again.

    Idempotency tokens must represent precisely one logical call to a webhook
    in the design of the state machine. For example:

    - An action being retried _must_ use the same idempotency token, in case
      the original failure was a network issue, and the receiver did indeed
      process the action.
    - An action being triggered again in a state machine that loops _must_ use
      a different token, as loops are a supported use-case.

    The label passed to this function _must_ be locked for the current
    transaction.
    """
    return hashlib.sha256(str(latest_history.id).encode('ascii')).hexdigest()



"""Helpers for working with timezones."""


import functools
from typing import FrozenSet

import dateutil.tz
import dateutil.zoneinfo


@functools.lru_cache(maxsize=1)
def get_known_timezones() -> FrozenSet[str]:
    """
    Return a cached set of the known timezones.

    This actually pulls its list from the internal database inside `dateutil`
    as there doesn't seem to be a nice way to pull the data from the system.

    These are expected to change sufficiently infrequently that this is ok. In
    any case, `dateutil` falls back to using this data source anyway, so at
    worst this is a strict subset of the available timezones.
    """
    # Get a non-cached copy of the `ZoneInfoFile`, not because we care about
    # the cache being out of date, but so that we're not stuck with a MB of
    # redundant memory usage.

    # ignore types because `dateutil.zoneinfo` isn't present in the typeshed
    info = dateutil.zoneinfo.ZoneInfoFile(  # type: ignore
        dateutil.zoneinfo.getzoneinfofile_stream(),  # type: ignore
    )

    return frozenset(info.zones.keys())



"""Shared types for state machine execution."""

from typing import Any, Dict, NamedTuple

Metadata = Dict[str, Any]


class LabelRef(NamedTuple):
    """API representation of a label for the state machine."""
    name: str
    state_machine: str



"""Processor classes to support cron scheduled jobs."""

import logging
import datetime
import functools
from typing import Any, List, Type, Callable

import dateutil.tz
from typing_extensions import Protocol

from routemaster.config import (
    TimezoneAwareTrigger,
    MetadataTimezoneAwareTrigger,
)
from routemaster.timezones import get_known_timezones
from routemaster.time_utils import time_appears_in_range
from routemaster.state_machine import (
    LabelProvider,
    labels_in_state_with_metadata,
)


def _logger_for_type(type_: Type[Any]) -> logging.Logger:
    return logging.getLogger(f"{type_.__module__}.{type_.__name__}")


class ProcessingSpecificCronProcessor(Protocol):
    """Type signature for the a processing-specific cron processor callable."""

    def __call__(
        self,
        *,
        label_provider: LabelProvider,
    ) -> None:
        """Type signature for a processing-specific cron processor callable."""
        ...


class TimezoneAwareProcessor:
    """
    Cron processor for the `TimezoneAwareTrigger`.

    This expects to be called regularly, but is tolerant of delays. It will
    only actually do any processing if the time for its trigger timezone has
    passed since it was last called (or constructed).

    Processing of delayed runs has two side-effects:
     - arbitrarily delayed processing will still run, but
     - delayed processing may cause multiple runs whose times have all passed
       between calls to be processed together as a single run
    """
    def __init__(
        self,
        processor: Callable[[], None],
        trigger: TimezoneAwareTrigger,
    ) -> None:
        self.processor = processor
        self.trigger = trigger
        self._last_call = datetime.datetime.now(dateutil.tz.tzutc())
        self._logger = _logger_for_type(type(self))

    def __call__(self) -> None:
        """Run the cron processing."""
        last_call = self._last_call
        self._last_call = now = datetime.datetime.now(dateutil.tz.tzutc())

        trigger_time = self.trigger.time.replace(
            tzinfo=dateutil.tz.gettz(self.trigger.timezone),
        )

        try:
            should_process = time_appears_in_range(
                when=trigger_time,
                start=last_call,
                end=now,
            )
        except ValueError:
            self._logger.exception(
                "Failed to determine whether trigger time has passed",
            )
            return

        if not should_process:
            self._logger.debug(
                f"Not currently time to do processing (waiting for "
                f"{self.trigger.time} in {self.trigger.timezone})",
            )
            return

        self._logger.info(
            f"Processing {self.trigger.time} in {self.trigger.timezone}",
        )
        self.processor()

    def __repr__(self) -> str:
        """Return a useful debug representation."""
        return (
            f'<TimezoneAwareProcessor: {self.trigger.time} in '
            f'{self.trigger.timezone}>'
        )


class MetadataTimezoneAwareProcessor:
    """
    Cron processor for the `MetadataTimezoneAwareTrigger`.

    This expects to be called regularly, but is tolerant of delays. It will
    only actually do any processing if its trigger time (for any known
    timezone) has pased since it was last called (or constructed). The
    processing it does is then filtered to labels whose timezone metadata is
    among the matched timezones.

    Processing of delayed runs has two side-effects:
     - arbitrarily delayed processing will still run, but
     - delayed processing may cause multiple runs whose times have all passed
       between calls to be processed together as a single run
    """
    def __init__(
        self,
        processor: ProcessingSpecificCronProcessor,
        trigger: MetadataTimezoneAwareTrigger,
    ) -> None:
        self.processor = processor
        self.trigger = trigger
        self._last_call = datetime.datetime.now(dateutil.tz.tzutc())
        self._logger = _logger_for_type(type(self))

    def __call__(self) -> None:
        """Run the cron processing."""
        last_call = self._last_call
        self._last_call = now = datetime.datetime.now(dateutil.tz.tzutc())

        try:
            timezones: List[str] = []
            for timezone in get_known_timezones():
                trigger_time = self.trigger.time.replace(
                    tzinfo=dateutil.tz.gettz(timezone),
                )
                matches = time_appears_in_range(
                    when=trigger_time,
                    start=last_call,
                    end=now,
                )
                if matches:
                    timezones.append(timezone)
        except ValueError:
            self._logger.exception(
                "Failed to determine whether trigger time has passed",
            )
            return

        if not timezones:
            self._logger.debug(
                f"Not currently time to do processing (waiting for "
                f"{self.trigger.time})",
            )
            return

        label_provider = functools.partial(
            labels_in_state_with_metadata,
            path=self.trigger.timezone_metadata_path,
            values=timezones,
        )

        self._logger.info(
            f"Processing {self.trigger.time} in {timezones} for "
            f"{self.trigger.timezone_metadata_path}",
        )
        self.processor(label_provider=label_provider)

    def __repr__(self) -> str:
        """Return a useful debug representation."""
        return (
            f'<MetadataTimezoneAwareProcessor: {self.trigger.time} for '
            f'{self.trigger.timezone_metadata_path}>'
        )



"""Periodic job running."""

import time
import functools
import itertools
import threading
from typing import Callable, Iterable

import schedule
from typing_extensions import Protocol

from routemaster.app import App
from routemaster.config import (
    Gate,
    State,
    Action,
    StateMachine,
    IntervalTrigger,
    MetadataTrigger,
    SystemTimeTrigger,
    TimezoneAwareTrigger,
    MetadataTimezoneAwareTrigger,
)
from routemaster.state_machine import (
    LabelProvider,
    LabelStateProcessor,
    process_cron,
    process_gate,
    process_action,
    labels_in_state,
    labels_needing_metadata_update_retry_in_gate,
)
from routemaster.cron_processors import (
    TimezoneAwareProcessor,
    MetadataTimezoneAwareProcessor,
)

IsTerminating = Callable[[], bool]


class CronProcessor(Protocol):
    """Type signature for the cron processor callable."""

    def __call__(
        self,
        *,
        app: App,
        state: State,
        state_machine: StateMachine,
        fn: LabelStateProcessor,
        label_provider: LabelProvider,
    ) -> None:
        """Type signature for the cron processor callable."""
        ...


class StateSpecificCronProcessor(Protocol):
    """Type signature for the a state-specific cron processor callable."""

    def __call__(
        self,
        *,
        fn: LabelStateProcessor,
        label_provider: LabelProvider,
    ) -> None:
        """Type signature for a state-specific cron processor callable."""
        ...


# The cron configuration works by building up a partially applied function
# `process_job`, at multiple levels. This allows us to hook an intermediate
# stage for testing.

def process_job(
    *,
    # Bound at the cron thread level
    is_terminating: IsTerminating,
    # Bound at the state scheduling level
    app: App,
    state: State,
    state_machine: StateMachine,
    # Bound when scheduling a specific job for a state
    fn: LabelStateProcessor,
    label_provider: LabelProvider,
):
    """Process a single instance of a single cron job."""

    def _iter_labels_until_terminating(
        state_machine: StateMachine,
        state: State,
    ) -> Iterable[str]:
        return itertools.takewhile(
            lambda _: not is_terminating(),
            label_provider(app, state_machine, state),
        )

    try:
        with app.logger.process_cron(state_machine, state, fn.__name__):
            process_cron(
                process=fn,
                get_labels=_iter_labels_until_terminating,
                app=app,
                state=state,
                state_machine=state_machine,
            )
    except Exception:
        return


def _configure_schedule_for_state(
    scheduler: schedule.Scheduler,
    processor: StateSpecificCronProcessor,
    state: State,
) -> None:
    if isinstance(state, Action):
        scheduler.every().minute.do(
            processor,
            fn=process_action,
            label_provider=labels_in_state,
        )
    elif isinstance(state, Gate):
        for trigger in state.triggers:
            if isinstance(trigger, SystemTimeTrigger):
                scheduler.every().day.at(
                    f"{trigger.time.hour:02d}:{trigger.time.minute:02d}",
                ).do(
                    processor,
                    fn=process_gate,
                    label_provider=labels_in_state,
                )
            elif isinstance(trigger, TimezoneAwareTrigger):
                func = functools.partial(
                    processor,
                    fn=process_gate,
                    label_provider=labels_in_state,
                )
                scheduler.every().minute.do(
                    TimezoneAwareProcessor(func, trigger),
                )
            elif isinstance(trigger, MetadataTimezoneAwareTrigger):
                scheduler.every().minute.do(
                    MetadataTimezoneAwareProcessor(
                        functools.partial(processor, fn=process_gate),
                        trigger,
                    ),
                )
            elif isinstance(trigger, IntervalTrigger):
                scheduler.every(
                    trigger.interval.total_seconds(),
                ).seconds.do(
                    processor,
                    fn=process_gate,
                    label_provider=labels_in_state,
                )
            elif isinstance(trigger, MetadataTrigger):  # pragma: no branch
                label_provider = labels_needing_metadata_update_retry_in_gate
                scheduler.every().minute.do(
                    processor,
                    fn=process_gate,
                    label_provider=label_provider,
                )
            else:
                # We only care about time based triggers and retries here.
                pass  # pragma: no cover
    else:
        raise RuntimeError(  # pragma: no cover
            f"Unsupported state type {state}",
        )


def configure_schedule(
    app: App,
    scheduler: schedule.Scheduler,
    processor: CronProcessor,
) -> None:
    """Set up all scheduled tasks that need running."""
    for state_machine in app.config.state_machines.values():
        for state in state_machine.states:
            _configure_schedule_for_state(
                scheduler,
                functools.partial(
                    processor,
                    app=app,
                    state=state,
                    state_machine=state_machine,
                ),
                state,
            )


class CronThread(threading.Thread):  # pragma: no cover
    """Background thread for running periodic jobs."""

    def __init__(self, app: App) -> None:
        self._terminating = False
        self.app = app
        self.scheduler = schedule.Scheduler()
        super().__init__(name="cron")

    def run(self) -> None:
        """Run main scheduling loop."""
        configure_schedule(
            self.app,
            self.scheduler,
            functools.partial(
                process_job,
                is_terminating=self.is_terminating,
            ),
        )
        self.app.logger.info("Starting cron thread")
        while not self.is_terminating():
            self.scheduler.run_pending()
            time.sleep(1)

    def stop(self) -> None:
        """Set the stopping flag and wait for thread end."""
        self._terminating = True
        self.app.logger.info("Cron thread shutting down")
        self.join()

    def is_terminating(self) -> bool:
        """Dynamically access whether we are terminating."""
        return self._terminating



"""Base class for logging plugins."""

import contextlib


class BaseLogger:
    """Base class for logging plugins."""

    def __init__(self, config, *args, **kwargs) -> None:
        self.config = config

    def init_flask(self, flask_app):
        """
        Entrypoint for configuring logging on the flask server.

        Note: this is only called if routemaster is being run as a server, not
        when validating configuration for example.
        """
        pass

    @contextlib.contextmanager
    def process_cron(self, state_machine, state, fn_name):
        """Wraps the processing of a cron job for logging purposes."""
        yield

    @contextlib.contextmanager
    def process_webhook(self, state_machine, state):
        """Wraps the processing of a webhook for logging purposes."""
        yield

    def process_request_started(self, environ):
        """Request started."""
        pass

    def process_request_finished(
        self,
        environ,
        status,
        headers,
        exc_info,
    ):
        """Completes the processing of a request."""
        pass

    def webhook_response(
        self,
        state_machine,
        state,
        response,
    ):
        """Logs the receipt of a response from a webhook."""
        pass

    @contextlib.contextmanager
    def process_feed(self, state_machine, state, feed_url):
        """Wraps the processing of a feed for logging purposes."""
        yield

    def feed_response(
        self,
        state_machine,
        state,
        feed_url,
        response,
    ):
        """Logs the receipt of a response from a feed."""
        pass

    def __getattr__(self, name):
        """Implement the Python logger API."""
        if name in (
            'debug',
            'info',
            'warning',
            'error',
            'critical',
            'exception',
        ):
            return self._log_handler

        raise AttributeError(name)

    def _log_handler(self, *args, **kwargs):
        pass



"""Core App singleton that holds state for the application."""
import threading
import contextlib
from typing import Dict, Optional

from sqlalchemy.orm import Session, sessionmaker
from sqlalchemy.engine import Engine

from routemaster.db import initialise_db
from routemaster.config import Config, StateMachine
from routemaster.logging import BaseLogger, SplitLogger, register_loggers
from routemaster.webhooks import (
    WebhookRunner,
    webhook_runner_for_state_machine,
)


class App(threading.local):
    """Core application state."""

    _db: Engine
    config: Config
    logger: BaseLogger
    _current_session: Optional[Session]
    _webhook_runners: Dict[str, WebhookRunner]

    def __init__(
        self,
        config: Config,
    ) -> None:
        """Initialisation of the app state."""
        self.config = config
        self.initialise()

    def initialise(self):
        """
        Initialise this instance of the app.

        This may be used to re-initialise after forking in a multiprocessing
        environment.
        """
        self._db = initialise_db(self.config.database)
        self.logger = SplitLogger(
            self.config,
            loggers=register_loggers(self.config),
        )
        self._sessionmaker = sessionmaker(self._db)
        self._current_session = None
        self._needs_rollback = False

        # Webhook runners may choose to persist a session, so we instantiate
        # up-front to ensure we re-use state.
        self._webhook_runners = {
            x: webhook_runner_for_state_machine(y)
            for x, y in self.config.state_machines.items()
        }

    @property
    def session(self) -> Session:
        """The current ORM session."""
        if self._current_session is None:
            raise RuntimeError(
                "There is no current session; you can only access `.session` "
                "within a `with app.new_session():` block.",
            )

        return self._current_session

    def set_rollback(self):
        """Mark the current session as needing rollback."""
        if self._current_session is None:
            raise RuntimeError(
                "There is no current session; you can only access `.session` "
                "within a `with app.new_session():` block.",
            )

        self._needs_rollback = True

    @contextlib.contextmanager
    def new_session(self):
        """Run a single session in this scope."""
        if self._current_session is not None:
            raise RuntimeError("There is already a session running.")

        self._current_session = self._sessionmaker()
        try:
            yield
            if self._needs_rollback:
                self._current_session.rollback()
            else:
                self._current_session.commit()

        except:  # noqa: B901, E722 - Unconditional re-raise
            self._current_session.rollback()
            raise

        finally:
            self._current_session.close()
            self._current_session = None
            self._needs_rollback = False

    def get_webhook_runner(self, state_machine: StateMachine) -> WebhookRunner:
        """Get the webhook runner for a state machine."""
        return self._webhook_runners[state_machine.name]
